class A:
    def __init__(self):
        self.name_of_var = 1
        self.a = 1
        self.b = 2
        self.c = 3
        self.d = 4
        self.e = 5
        print(dict(self.__dict__.items()))


a = A()
d = {"A": 0, "B": 1, "C":2}
for key, arg in dict(zip(list(a.__dict__.keys())[:2],
                                 list(a.__dict__.values())[:2])).items():
    print(key, arg)

print("OK")







print(dict(zip(list(d.values())[:1], list(d.keys())[:1])))



import socket
import sys

import pygame.time

from pyreverb.Exemple.shooter_objects import Player, TICK
from pyreverb.reverb import ReverbManager, ReverbSide, PATH_LOG
from pyreverb.reverb_kernel import server_event_registry, Server, save_logs

clock = pygame.time.Clock()


@server_event_registry.on_event("client_connection")
def on_connecting(clt: socket.socket, *args):
    ReverbManager.add_new_reverb_object(
        Player(pos=[400, 400], color=Player.choose_rnd_color(), belonging_membership=clt.getpeername()[1]))


@server_event_registry.on_event("client_disconnection")
def on_disconnecting(clt: socket.socket, *args):
    for p in ReverbManager.get_all_ro_by_type(Player):
        if p.belonging_membership == clt.getpeername()[1]:
            ReverbManager.remove_reverb_object(p.uid)


def start_server(port: int, admin_key: int):
    print("Starting server...")
    ReverbManager.REVERB_SIDE = ReverbSide.SERVER
    serv = Server(port=port)
    ReverbManager.REVERB_CONNECTION = serv
    ReverbManager.ADMIN_KEY = admin_key  # Set the admin key
    serv.start_server()
    while True:
        try:
            clock.tick(TICK)
            ReverbManager.server_sync()
        except KeyboardInterrupt:
            serv.stop_server()
            break
        except Exception as e:
            print("ERROR CLOSE")
            save_logs(PATH_LOG)
            raise e


if __name__ == "__main__":
    print("Starting server from main!")
    port = 8080
    if len(sys.argv) > 1:
        port = int(sys.argv[1])

    admin_key = 1000
    if len(sys.argv) > 2:
        admin_key = int(sys.argv[2])
    start_server(port, admin_key)



import atexit
import platform
import random
import subprocess
import uuid
from enum import Enum
from typing import Type, TypeVar

from .reverb_errors import *
from .reverb_kernel import *

T = TypeVar("T")

VERBOSE = 2
WORKING_DIR = os.path.dirname(os.path.abspath(__file__))
PATH_LOG = f"{WORKING_DIR}/logs/"

"""
- 2: Full verbose
- 1: Only some things
- 0: Stop verbosing
"""


class ReverbSide(Enum):
    SERVER = 1
    CLIENT = 2


def start_distant(file, *args, **kwargs) -> subprocess.Popen:
    """
    Sart a process of the game with his side.
    :param file: The name of the file to be executed
    :param args: More argues
    :param kwargs: More dict arguments
    :return: The process
    """
    system = platform.system()
    if system == "Windows":
        print("Starting distant file on Windows")
        return subprocess.Popen([sys.executable, file] + list(args) + list(kwargs),
                                creationflags=subprocess.CREATE_NEW_CONSOLE)
    elif system == "Linux":
        terms = [
            ["qterminal", "-e"],
            ["gnome-terminal", "--"],
            ["konsole", "-e"],
            ["xterm", "-e"],
        ]
        for term in terms:
            try:
                return subprocess.Popen(term + ["python3", file] + list(args) + list(kwargs))
            except FileNotFoundError:
                continue
        raise RuntimeError(f"No terminal found: {[t[0] for t in terms]}")

    else:
        raise OSError("Unsupported OS!")


def stop_subprocess(sub: subprocess.Popen = None):
    """
    Stop a process
    DON'T SAVE LOG IF SERVER IS CLOSED ON WINDOWS WITH THIS METHODE!!!
    :param sub: The subprocess
    """
    if sub and sub.poll() is None:
        sub.terminate()
        sub.wait()
        print("Server Closed!")
    else:
        warn(
            "You try to stop the process with a None subprocess! By default the subprocess is reverb.SERVER_PROCESS, set it to the server process before run it again if args is let by default.")


class SyncVar:
    """
    Simple class that trigger a hook when the value changes and syncs var between all clients
    """

    def __init__(self, default=None, on_changed: list[staticmethod] = None):
        """
        :param default: The value
        :param on_changed: List of methode that will be trigger if the value that will be set changes
        """
        if on_changed is None:
            on_changed = []
        self.on_changed = on_changed
        self.value = default
        self.has_changed = False

    def get(self, val_if_not_found=None, get_only_if_change=False) -> object:
        """
        Get the value
        :param val_if_not_found: Default value that will be return if the value is None
        :param get_only_if_change: Will return None if the value is not previously change
        (Careful it is not really sync, it is just for Reverb)
        :return: The value
        """
        if get_only_if_change and not self.has_changed:
            return None
        return self.value if self.value else val_if_not_found

    def set(self, val):
        """
        Set a value
        :param val: The value
        """
        old = self.value
        self.value = val
        if old != val:
            self.has_changed = True
            for func in self.on_changed:
                func(old, val)


def check_if_json_serializable(*args: SyncVar):
    for arg in args:
        try:
            json.dumps(arg)
        except (TypeError, OverflowError):
            raise Exception(
                f"The arg: {arg} is not serializable ! It has to be serializable by JSON to be agree as a reverb_args.")


class ReverbObject:
    """
    - Base class of all object connected to the Network
    """

    def __init__(self, *reverb_args: SyncVar, uid: str = "Unknown", belonging_membership: int = None):
        """
        :param reverb_args: All the custom vars
        :param uid: The uid of the object, let it on None if you are not sure of what you're doing here
        :param belonging_membership: This refers to the port of a client. With this you can know if the RO is from a local instance or not. Let it on None if you're not sure what you're doing here
        """
        self.belonging_membership = belonging_membership
        self.reverb_args = reverb_args
        self.uid: str = uid
        self.is_alive = True
        self.type = self.__class__.__name__
        self.is_initialized = False

    def get_sync_vars(self, get_value=False, get_only_if_changed=True) -> dict[str, SyncVar | object]:
        """
        List all SyncVars initialized into the ReverbObject
        :param get_value: Will get the value of the SyncVar if True else will return the object
        :param get_only_if_changed: Get the value only if changed
        :return: A dict of all SyncVars or val of the SyncVar by their name
        """
        sync_vars = {}
        for key, arg in dict(zip(list(self.__dict__.keys())[:len(self.reverb_args)],
                                 list(self.__dict__.values())[:len(self.reverb_args)])).items():
            if isinstance(arg, SyncVar):
                val = (arg.get(get_only_if_changed) if get_value else arg)
                if get_only_if_changed:
                    if arg.has_changed:
                        sync_vars[key] = val
                else:
                    sync_vars[key] = val
                if get_only_if_changed:
                    arg.has_changed = False
        return sync_vars

    def pack(self, only_sync_vars) -> list[object]:
        """
        :param only_sync_vars: If True, it does not return a type + belonging_membership but just sync_vars
        :return: A dict of all necessary args that are linked between the server and the clients by their name in the class
        """
        sync_vars = self.get_sync_vars(get_value=True, get_only_if_changed=only_sync_vars)
        check_if_json_serializable(*[sync_vars])

        # If not init yet: send the type and the belonging_membership to construct the object -> if no sync vars pack nothing ->
        return ([self.type, self.belonging_membership] if not only_sync_vars else []) + [sync_vars]

    def sync(self, reverb_args: dict[str, SyncVar]):
        """
        - Call on the 'CLIENT' side to sync new ro data
        :param reverb_args: List of args to be updated
        """
        if ReverbManager.REVERB_SIDE == ReverbSide.CLIENT:
            for key, val in reverb_args.items():
                getattr(self, key).set(val)
            self.reverb_args = list(reverb_args.values())
        else:
            raise ReverbWrongSideError(ReverbManager.REVERB_SIDE.name)

    @staticmethod
    def print_object(msg):
        """
        - Print a message with the ReverbObject style
        :param msg: The message
        """
        print(f"{Back.MAGENTA + Fore.RED}[{Fore.RESET}REVERB_OBJECT{Fore.RED}]{Style.RESET_ALL} {msg}")

    def is_owner(self) -> bool:
        """
        - Chek if the ReverbObject is a membership of this client
        - Only call on the 'CLIENT' side
        :return:
        """
        if ReverbManager.REVERB_SIDE == ReverbSide.CLIENT:
            return ReverbManager.REVERB_CONNECTION.client.getsockname()[1] == self.belonging_membership
        else:
            raise ReverbWrongSideError(ReverbManager.REVERB_SIDE)

    def compute_server(self, func, *args):
        """
        - Send a Packet to the server to compute a function server with args
        - Only on 'CLIENT' side
        :param func: The server function reference. Has to be into the Class
        :param args: Args of the function
        """
        if self.is_alive:
            ReverbManager.REVERB_CONNECTION.send("calling_server_computing", self.uid, func.__name__, *args)

    def compute_client(self, func, *args):
        """
        - Send a Packet to the client to compute a function client with args
        - Only on 'SERVER' side
        :param func: The client function reference. Has to be into the Class
        :param args: Args of the function
        """
        if self.is_alive:
            ReverbManager.REVERB_CONNECTION.send("calling_client_computing", self.uid, func.__name__, *args)

    def is_uid_init(self) -> bool:
        """
        :return: if uid is an init or not
        """
        return self.uid != "Unknown"

    def on_init_from_client(self):
        """
        - Call on the 'CLIENT' side
        - Override this function
        - Call when the object is creating from the 'Client' side
        """

    def on_init_from_server(self):
        """
        - Call on 'SERVER' side
        - Override this function
        - Call when the object is creating from the 'Server' side
        """

    def on_destroy_from_client(self):
        """
        - Call on the 'CLIENT' side
        - Override this function
        - Call when the object is removing from the 'CLIENT' side
        """

    def on_destroy_from_server(self):
        """
        - Call on 'SERVER' side
        - Override this function
        - Call when the object is removing from the 'SERVER' side
        """

    def __del__(self):
        if VERBOSE == 2:
            ReverbObject.print_object(f"Destroying the object {self.uid=}")


class ReverbManager:
    """
    - This class is static!
    - It links ReverbObject to the reference of the ReverbObject!
    """
    REVERB_SIDE: ReverbSide = None
    REVERB_CONNECTION: Client | Server = None  # Client, or Server
    REVERB_OBJECTS: dict[str, ReverbObject] = {}
    REVERB_OBJECT_REGISTRY = {"ReverbObject": ReverbObject}  # Register all type
    ADMIN_KEY = random.randint(1000, 10000)
    ADMINS = []

    try:
        IS_HOST = sys.argv[2] == "1"
        """Set to true automatically if is_host param passed as param otherwise False"""
    except:
        IS_HOST = False  # Check if host

    @staticmethod
    def print_manager(msg):
        """
        - Print a message with the ReverbManager style
        :param msg: The message
        """
        if VERBOSE != 0:
            print(f"{Back.YELLOW + Fore.RED}[{Fore.RESET}REVERB_MANAGER{Fore.RED}]{Style.RESET_ALL} {msg}")

    @staticmethod
    def add_type_if_dont_exit(ro: type[ReverbObject]):
        try:
            ReverbManager.REVERB_OBJECT_REGISTRY[ro.__name__]
        except KeyError:
            ReverbManager.REVERB_OBJECT_REGISTRY[ro.__name__] = ro
            if VERBOSE >= 1:
                ReverbManager.print_manager(f"Adding type '{ro.__name__}' to the registry.")

    @staticmethod
    @server_event_registry.on_event("stop_server_admin")
    def on_stop_server_admin(clt: socket.socket):
        """
        - Call on the 'SEVER' side
        - Shutdown the server!
        """
        if ReverbManager.REVERB_SIDE == ReverbSide.SERVER:
            if clt in ReverbManager.ADMINS:
                ReverbManager.print_manager("An admin stop the server!")
                ReverbManager.REVERB_CONNECTION.stop_server()
                save_logs(PATH_LOG)
                os._exit(0)
            else:
                ReverbManager.print_manager("A user tried to stop the server without admin rights!")
        else:
            raise ReverbWrongSideError(ReverbSide.SERVER)

    @staticmethod
    def stop_server_admin():
        """
        - Call on the 'CLIENT' side
        - Send a package to stop the server. Only stop if the client is registered as an admin.
        """
        if ReverbManager.REVERB_SIDE == ReverbSide.CLIENT:
            ReverbManager.REVERB_CONNECTION.send("stop_server_admin")
        else:
            raise ReverbWrongSideError(ReverbSide.CLIENT)

    @staticmethod
    def log_as_admin(key: int):
        """
        - Call on the 'CLIENT' side only
        - Try to grant admin right to the client
        """
        if ReverbManager.REVERB_SIDE == ReverbSide.CLIENT:
            ReverbManager.REVERB_CONNECTION.send("grant_admin", key)
        else:
            raise ReverbWrongSideError(ReverbSide.CLIENT)

    @staticmethod
    @server_event_registry.on_event("grant_admin")
    def on_grant_admin(clt: socket.socket, key):
        """
        - Call on the 'SERVER' side
        - Grant admin right to a client
        :param key: Admin key sent by the client
        """
        response = "REFUSED"
        if key == ReverbManager.ADMIN_KEY and clt not in ReverbManager.ADMINS:
            ReverbManager.ADMINS.append(clt)
            ReverbManager.print_manager(f"The user '{clt.getpeername()[0]}' has been successfully granted to admin!")
            response = "GRANTED"
        else:
            ReverbManager.print_manager(
                f"WARNING: The user '{clt.getpeername()[0]}' tried to connect as admin but with wrong ADMIN_KEY or is already an admin!")
        ReverbManager.REVERB_CONNECTION.send_to(clt, "grant_admin_response", response)

    @staticmethod
    @client_event_registry.on_event("grant_admin_response")
    def on_grant_admin_response(clt: socket.socket, response):
        if response == "GRANTED":
            ReverbManager.print_manager(f"Successfully granted to admin!")
        elif response == "REFUSED":
            ReverbManager.print_manager(f"The server refused the admin right to you! (Wrong key or already admin)")

    @staticmethod
    def server_sync():
        """
        - Call on 'SERVER' side
        - Sync value from 'SERVER' to 'CLIENT' side
        """
        if ReverbManager.REVERB_SIDE == ReverbSide.SERVER:
            ros = {}
            does_something_changed = False
            # Avoiding: "RuntimeError: dictionary changed size during iteration"
            for uid, ro in list(ReverbManager.REVERB_OBJECTS.items()):
                if ro != "DESTROYED":
                    pack = ro.pack(only_sync_vars=ro.is_initialized)
                    if pack and pack != [{}]:
                        ros[uid] = pack
                        does_something_changed = True
                    if not ro.is_initialized:
                        ro.is_initialized = True
            if does_something_changed:
                ReverbManager.REVERB_CONNECTION.send_to_all("server_sync", ros)
        else:
            raise ReverbWrongSideError(ReverbManager.REVERB_SIDE)

    @staticmethod
    @server_event_registry.on_event("client_connection")
    def on_client_connect(clt: socket.socket, *args):
        """
        - 'Server' side
         - Spawn existent ro on the new client!
        """
        ros = {}
        for uid, ro in list(ReverbManager.REVERB_OBJECTS.items()):
            if ro != "DESTROYED":
                pack = ro.pack(only_sync_vars=False)
                if pack:
                    ros[uid] = pack
        ReverbManager.REVERB_CONNECTION.send_to(clt, "server_sync", ros)

    @staticmethod
    def get_reverb_object(uid: str) -> ReverbObject:
        """
        - Get the reverb object by uid
        :param uid: The uid
        :return: ReverbObject or ReverbObjectNotFoundError if not found
        """
        try:
            return ReverbManager.REVERB_OBJECTS[uid]
        except KeyError:
            raise ReverbObjectNotFoundError(uid)

    @staticmethod
    def get_cls_by_type_name(t: str):
        try:
            return ReverbManager.REVERB_OBJECT_REGISTRY[t]
        except KeyError:
            raise ReverbTypeNotFoundError(t)

    @staticmethod
    def get_all_ro_by_type(t: Type[T]) -> list[T]:
        """
        - Get all the ReverbObject by a type
        :param t: Type of ReverbObject
        :return: Return the list of all found same types into the ReverbManager
        """
        ros = []
        for uid, ro in ReverbManager.REVERB_OBJECTS.items():
            if ro != "DESTROYED":
                if isinstance(ro, t):
                    ros.append(ro)
        return ros

    @staticmethod
    def add_new_reverb_object(ro: ReverbObject):
        """
        - Add a new ReverbObject to the ReverbManager
        :param ro: The ReverbObject
        """
        if ro not in ReverbManager.REVERB_OBJECTS.values():  # Check if the
            if ReverbManager.REVERB_SIDE == ReverbSide.SERVER:  # check RM side
                if not ro.is_uid_init():  # Check if the RO is not init yet
                    # SERVER
                    uid = str(uuid.uuid4())
                    ReverbManager.REVERB_OBJECTS[uid] = ro
                    ro.uid = uid
                    threading.Thread(target=ro.on_init_from_server, daemon=True).start()
                else:
                    raise ReverbUIDAlreadyInitError(ro, ro.uid)
            else:
                # CLIENT
                if ro.is_uid_init():
                    ReverbManager.REVERB_OBJECTS[ro.uid] = ro
                    ro.is_initialized = True
                else:
                    raise ReverbUIDUnknownError()
                threading.Thread(target=ro.on_init_from_client, daemon=True).start()
        else:
            raise ReverbObjectAlreadyExistError(ro)
        if VERBOSE == 2:
            ReverbManager.print_manager(
                f"New ReverbObject: {ro} add into '{ReverbManager.REVERB_SIDE.name}' side with uid={ro.uid}")

    @staticmethod
    def remove_reverb_object(uid: str):
        """
        - Call on 'SERVER' side only
        - Remove the ro from all clients
        :param uid: The uid of the RO
        """
        if ReverbManager.REVERB_SIDE == ReverbSide.SERVER:
            try:
                ro: ReverbObject = ReverbManager.get_reverb_object(uid)
                ro.is_alive = False

                threading.Thread(target=ro.on_destroy_from_server, daemon=True).start()
                ReverbManager.REVERB_OBJECTS[uid] = "DESTROYED"
                f = lambda: (time.sleep(3), ReverbManager.REVERB_OBJECTS.pop(
                    uid))  # Remove the ro 3 sec after on the server to avoid syncing bugs
                threading.Thread(target=f, daemon=True).start()
            except KeyError:
                raise KeyError(f"The {uid=} is not found !")

            ReverbManager.REVERB_CONNECTION.send_to_all("remove_ro", uid)
        else:
            raise ReverbWrongSideError(ReverbManager.REVERB_SIDE)

    @staticmethod
    @client_event_registry.on_event("remove_ro")
    def on_server_remove_reverb_object(clt: socket.socket, uid, *args):
        """
        - Only call on 'CLIENT' side
        - Remove the object
        :param clt: The socket
        :param uid: The uid of the ReverbObject to delete
        """
        if ReverbManager.REVERB_SIDE == ReverbSide.CLIENT:
            ro: ReverbObject = ReverbManager.get_reverb_object(uid)
            ro.is_alive = False
            ReverbManager.REVERB_OBJECTS.pop(uid)
            threading.Thread(target=ro.on_destroy_from_client(), daemon=True).start()
        else:
            raise ReverbWrongSideError(ReverbManager.REVERB_SIDE)

    @staticmethod
    @client_event_registry.on_event("server_sync")
    def on_server_sync(clt: socket.socket, ros: dict[str, list[int | str | dict[str, object]]], *args):
        """
        - Called on the 'Client' side
        - Called when the server syncs the state of ReverbObject with clients
        :param clt: The client socket
        :param ros: Dict[uids: list[list(values)]]
        """
        for uid, ro_data in ros.items():

            ro: ReverbObject = None

            try:  # try to get a reverb_object
                ro = ReverbManager.get_reverb_object(uid)
                ro_data = ro_data[0]
            except ReverbObjectNotFoundError:  # create a new one
                t: str = ro_data[0]  # Type
                cls = ReverbManager.get_cls_by_type_name(t)  # Class
                args = list(ro_data[2].values())  # arguments

                try:
                    ro = cls(*args, belonging_membership=ro_data[1])
                except TypeError:
                    raise TypeError(
                        f"Not enough param passed! You try to construct {cls} but those elements are passed {args}, {ro_data}")
                ro.uid = uid
                ReverbManager.add_new_reverb_object(ro)
                ro_data = ro_data[2]
            ro.sync(ro_data)

    @staticmethod
    @server_event_registry.on_event("calling_server_computing")
    def on_calling_server_computing(clt: socket.socket, uid: str, func_name: str, *args):
        """
        - Called on the 'Server' side
        - Called when a ReverbObject send data to be computed by the server (like movements, interactions, etc.)
        :param clt: The client socket
        :param uid: The uid of the ReverbObject
        :param func_name: The function name
        :param args: Params of the function
        """
        try:
            ro = ReverbManager.get_reverb_object(uid)
            if ro == "DESTROYED":
                return
        except ReverbObjectNotFoundError:
            warn(f"You try to compute on the server with a uid not found {uid=}.\n"
                 f"This may occur because the ro was removed and the syncing between the client and the server is not enough fast! or just because the uid is real2"
                 f"ly not found!")
            return

        try:
            func = getattr(ro, func_name)
            if args == ():
                func()
            else:
                func(*args)
        except AttributeError:
            raise NameError(f"The {func_name=} wasn't found into the ReverbObject!")

    @staticmethod
    @client_event_registry.on_event("calling_client_computing")
    def on_calling_client_computing(clt: socket.socket, uid: str, func_name: str, *args):
        """
        - Called on the 'Client' side
        - Called when a ReverbObject send data to be computed by the client
        :param clt: The socket
        :param uid: The uid of the ReverbObject
        :param func_name: The function name
        :param args: Params of the function
        """
        try:
            ro = ReverbManager.get_reverb_object(uid)
            if ro == "DESTROYED":
                return
        except ReverbObjectNotFoundError:
            warn(f"You try to compute on the client with a uid not found {uid=}.\n"
                 f"This may occur because the ro was removed and the syncing between the client and the server is not enough fast! or just because the uid is rea"
                 f"ly not found!")
            return

        try:
            func = getattr(ro, func_name)
            if args == ():
                func()
            else:
                func(*args)
        except AttributeError:
            raise NameError(f"The {func_name=} wasn't found into the ReverbObject!")

    @staticmethod
    def reverb_object_attribute(cls):
        """
        - Decorator of a ReverbObject class and add the type into the ReverbManager
        :param cls: The class
        :return: cls
        """
        if issubclass(cls, ReverbObject):
            ReverbManager.add_type_if_dont_exit(cls)
        else:
            raise TypeError(f"The class {cls} must be derivative from a ReverbObject!")
        return cls


def handle_exit():
    """
    Trigger on exit
    """
    if ReverbManager.REVERB_SIDE == ReverbSide.SERVER:
        save_logs(PATH_LOG)


atexit.register(handle_exit)



import random
import time

import pygame.time
from pygame import Vector2

from pyreverb.reverb import ReverbObject, ReverbManager, SyncVar

TICK = 60
MAP_SIZE = (800, 800)

clock = pygame.time.Clock()

@ReverbManager.reverb_object_attribute
class Bullet(ReverbObject):
    def __init__(self, pos, dir, color, belonging_membership: int = None):
        print(pos, dir, color)
        self.pos = SyncVar(pos)
        self.dir = SyncVar(dir)
        self.color = SyncVar(color)
        self.speed = 2
        super().__init__(self.pos, self.dir, self.color, belonging_membership=belonging_membership)

    def on_init_from_client(self):
        if self.is_owner():
            self.compute_server(self.die_after_time, 2)
            self.compute_server(self.update)

    # SERVER SIDE
    def die_after_time(self, t):
        time.sleep(t)
        ReverbManager.remove_reverb_object(self.uid)

    def update(self):
        while self.is_alive:
            self.pos.set(list(self.pos.get() + Vector2(self.dir.get()) * self.speed))
            clock.tick(TICK)


@ReverbManager.reverb_object_attribute
class Player(ReverbObject):
    def __init__(self, pos=[0, 0], dir=[0, 0], color="red", belonging_membership: int = None):
        self.pos = SyncVar(pos)
        self.dir = SyncVar(dir)
        self.color = SyncVar(color)
        super().__init__(self.pos, self.dir, self.color, belonging_membership=belonging_membership)

    def on_init_from_client(self):
        while self.is_alive:
            if self.is_owner():
                keys = pygame.key.get_pressed()
                dir = ""
                if keys[pygame.K_z]:
                    dir += "Z"
                if keys[pygame.K_s]:
                    dir += "S"
                if keys[pygame.K_q]:
                    dir += "Q"
                if keys[pygame.K_d]:
                    dir += "D"

                if dir != "":
                    self.compute_server(self.check_walk, dir)

                if keys[pygame.K_SPACE]:
                    self.compute_server(self.spawn_bullet)
                    time.sleep(1)

                clock.tick(TICK)

    @staticmethod
    def choose_rnd_color():
        return ["green", "red", "blue", "yellow"][random.randint(0, 3)]

    # ON SERVER
    def check_walk(self, dir):
        self.dir.set([0, 0])
        speed = 5

        def is_pos_in_map_bound(pos: Vector2):
            return 0 <= pos.x <= MAP_SIZE[0] and 0 <= pos.y <= MAP_SIZE[1]

        for d in dir:
            l_pos = {"Z": (0, -1), "S": (0, 1), "D": (1, 0), "Q": (-1, 0)}
            self.dir.set(tuple(self.dir.get() + Vector2(l_pos[d])))

        new_pos = self.pos.get() + Vector2(self.dir.get()) * speed
        if is_pos_in_map_bound(new_pos):
            self.pos.set(tuple(new_pos))

    def spawn_bullet(self):
        ReverbManager.add_new_reverb_object(
            Bullet(self.pos.get(), self.dir.get(), self.color.get(), belonging_membership=self.belonging_membership))



import datetime
import json
import os
import re
import socket
import struct
import sys
import threading
import time
from io import StringIO
from json import JSONDecodeError
from warnings import warn

from colorama import Fore, Back, Style


class Tee:
    """Store the console"""
    ANSI_ESCAPE = re.compile(r'\x1B\[[0-?]*[ -/]*[@-~]')

    def __init__(self, *streams, log_buffer=None):
        self.streams = streams
        self.log_buffer = log_buffer or StringIO()
        self._buffer = ""

    def write(self, data):
        for s in self.streams:
            s.write(data)
            s.flush()
        clean_data = Tee.ANSI_ESCAPE.sub('', data)

        self._buffer += clean_data
        while '\n' in self._buffer:
            line, self._buffer = self._buffer.split('\n', 1)
            timestamp = datetime.datetime.now().strftime('%H:%M:%S:%f')
            self.log_buffer.write(f"[{timestamp}] | {line}\n")
            self.log_buffer.flush()

    def flush(self):
        # Pour finir la dernière ligne si pas de \n
        if self._buffer:
            timestamp = datetime.datetime.now().strftime('%H:%M:%S:%f')
            self.log_buffer.write(f"[{timestamp}] | {self._buffer}\n")
            self.log_buffer.flush()
            self._buffer = ""
        for s in self.streams:
            s.flush()


# Storing output into a buffer
log_buffer = StringIO()
sys.stdout = Tee(sys.__stdout__, log_buffer=log_buffer)
sys.stderr = Tee(sys.__stderr__, log_buffer=log_buffer)


def save_logs(path: str = "./logs/"):
    file_path = f"{path}/log-{datetime.datetime.now().strftime('%y-%m-%d-%H-%M-%S')}.log"

    os.makedirs(os.path.dirname(file_path), exist_ok=True)
    print("Saving logs from this process...")
    with open(file_path, "w", encoding="utf-16") as log_file:
        log_file.write(log_buffer.getvalue())
        log_file.write(
            "\n\n\nDue to the stoping of the distant server (if it is the case), sometime logs can have error at the end!\n"
            "It is normal because when the server stop some threads are continuously opened.\n"
            "Closed the server make them crash! BUT IT IS NORMAL (normally normal it depends sometimes it is bad xd)\n"
            "Nevertheless examined them is a good things (you can have real errors!)")
        log_file.close()
    print("Log saved!")


class EventRegistry:
    """
    A Class that store events and handle them!
    """

    def __init__(self):
        self._events = {}

    def add_event(self, func, event_name):
        """
        Add an event to the EventRegistry
        :param func: The function
        :param event_name: The name of the event
        """
        if event_name not in self._events:
            self._events[event_name] = []
        self._events[event_name].append(func)

    def on_event(self, event_name):
        """
        Simple decorator to trigger events
        :param event_name: The name of the event
        :return: The decorator
        """

        def decorator(func):
            self.add_event(func, event_name)
            return func

        return decorator

    def get(self, event_name):
        """
        Get an event by his name
        :param event_name: The name of the event
        :return: Event functions
        """
        return self._events.get(event_name)

    def trigger(self, event_name, sock, *args, threading_event=True):
        """
        Trigger an event
        :param sock: the reference of the outcoming socket packet's
        :param event_name: The name of the event
        :param threading_event: If true, it will handle the event into a new thread else it will just execute the event into the main thread
        """
        handlers = self._events.get(event_name, [])  # Check if the event name contains functions or not
        if handlers:
            for handler in handlers:
                try:
                    if threading_event:
                        threading.Thread(target=handler, args=(sock, *args), daemon=True).start()
                    else:
                        handler(sock, *args)
                except TypeError:
                    if threading_event:
                        threading.Thread(target=handler, args=sock, daemon=True).start()
                    else:
                        handler(sock)
        else:
            warn(f"The handler for '{event_name}' is not found ! It may be normal, ignore then.")

    def all_events(self):
        """
        :return: All events
        """
        return list(self._events.keys())


client_event_registry = EventRegistry()
server_event_registry = EventRegistry()


class Packet:
    """
    Manager of packets
    """

    @staticmethod
    def create_packet(name: str, *content):
        """
        Create a packet and encode it
        :param name: Name of the packet/event
        :param content: The contents to send
        :return: An encoded packet ready to be sent :)
        """
        return json.dumps({"name": name, "contents": content}).encode()

    @staticmethod
    def recv_exact(sock: socket.socket, n: int):
        data = b""
        try:
            while len(data) < n:
                chunk = sock.recv(n - len(data))
                if not chunk:
                    raise ConnectionError("The socket is close...")
                data += chunk
            return data
        except ConnectionAbortedError:
            return b""

    @staticmethod
    def decode_packet(packet: bytes):
        """
        Decode the packet from a byte
        :param packet: The encoded packet
        :return: The name/event and the contents
        """
        try:
            decoded_packet = json.loads(packet.decode())
            return decoded_packet["name"], decoded_packet["contents"]
        except JSONDecodeError:
            warn(f"An error occurred with this packet: {packet.decode()}")
        except KeyError:
            warn(f"The packet is not valid ! A valid packet must have a 'name' and a 'contents' argument !")


class Client:
    """
    - A class that connect to a Server
    """

    def __init__(self, ip="127.0.0.1", port=8080):
        """
        :param ip: Ip server's
        :param port: Port server
        """
        self.port = port
        self.ip = ip
        self.client: socket.socket = None
        self.is_connected = False

    def connect(self):
        """
        Call to connect to the server
        """
        self.client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            self.client.connect((self.ip, self.port))
            self.is_connected = True

            threading.Thread(target=self.listen, daemon=True).start()
            client_event_registry.trigger("connection", self.client)  # Trigger connection event
        except ConnectionRefusedError:
            Client.print_client("The server is unreachable !")
        except socket.gaierror:
            Client.print_client("Error with host name or IP unfound")
        except TimeoutError:
            Client.print_client("Connexion TimeOut !")

    def listen(self):
        """
        Thread that listens for new content from the server
        """
        try:
            while self.is_connected:
                try:
                    raw_len = Packet.recv_exact(self.client, 4)  # Get the length of the packet
                    length = struct.unpack("!I", raw_len)[0]
                    packet = Packet.recv_exact(self.client, length)
                    if packet:
                        packet_name, contents = Packet.decode_packet(packet)
                        if packet_name == "server_stop":
                            client_event_registry.trigger(packet_name, self.client,
                                                          *contents)  # Trigger the event linked to the message of the server
                            Client.print_client("Server stopped !")
                            break
                        client_event_registry.trigger(packet_name, self.client,
                                                      *contents)  # Trigger the event linked to the message of the server
                    else:
                        Client.print_client("The server send an empty packet ! Closing...")
                        break
                except ConnectionResetError:
                    Client.print_client("Connection lost !")
                    break
                except Exception as e:
                    raise Exception(f"THIS IS NOT NORMAL:\n{e}")

        finally:
            self.disconnect()

    def send(self, packet_name: str, *content):
        """
        Send a content to the server
        :param packet_name: The name of the packet
        :param content: contents
        """
        if self.is_connected:
            packet = Packet.create_packet(packet_name, *content)
            length = len(packet)
            header = struct.pack('!I', length)
            try:
                self.client.sendall(header + packet)
            except BrokenPipeError:
                warn("The client has been disconnected during a sending operation!")
            except ConnectionResetError:
                warn("Server close or client disconnected during a sending operation!")
            except Exception as e:
                raise Exception(f"THIS IS NOT NORMAL DURING A SEND OPERATION:\n{e}")

    def disconnect(self):
        """
        Call to disconnect the user
        """
        if self.is_connected:
            try:
                self.send("client_disconnection", self.client.getpeername())
            finally:
                self.is_connected = False
                client_event_registry.trigger("disconnection", self.client)
                Client.print_client("Client close and disconnect from the server !")
                self.client.close()  # Close the client

    @staticmethod
    def print_client(msg):
        """
        Print a message with client style
        :param msg: the message to print
        """
        print(f"{Back.BLUE + Fore.RED}[{Fore.RESET}CLIENT{Fore.RED}]{Style.RESET_ALL} {msg}")


class Server:
    """
    - A class that open a Server
    """

    def __init__(self, host="", port=8080, ):
        """
        :param host: The ip. Let it him by default
        :param port: The listen port!
        """
        self.host = host
        self.port = port
        self.server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.is_online = False
        self.clients: dict[tuple[str, int], socket.socket] = {}

    @staticmethod
    def print_server(msg):
        """
        Print a message with server style
        :param msg: the message to print
        """
        print(f"{Back.GREEN + Fore.RED}[{Fore.RESET}SERVER{Fore.RED}]{Style.RESET_ALL} {msg}")

    def start_server(self):
        """
        Start the server
        """
        Server.print_server("Starting server...")
        self.server.bind(("", self.port))
        self.server.listen()

        Server.print_server(f"Server online ! Waiting for clients on {self.host}:{self.port}...")
        self.is_online = True
        threading.Thread(target=self._accept_clients, daemon=True).start()

    def stop_server(self):
        """
        Stop the server
        """
        self.is_online = False
        clts = list(self.clients.values())  # To avoid bugs
        self.send_to_all("server_stop")
        for client in clts:
            Server.print_server(f"The client: {client.getpeername()} is disconnect !")
            client.close()
        Server.print_server("All clients disconnected.")

        if self.server:
            self.server.close()
        Server.print_server("Server closed !")

    def _accept_clients(self):
        """Thread that accepts clients when they connect to the server"""
        try:
            while self.is_online:
                client_socket, addr = self.server.accept()
                self.clients[addr] = client_socket
                server_event_registry.trigger("client_connection", client_socket)
                threading.Thread(target=self._handle_client, args=(client_socket, addr), daemon=True).start()
        except KeyboardInterrupt:
            self.stop_server()
        except OSError:
            pass
        finally:
            Server.print_server("Server stop listening to new clients !")

    def _handle_client(self, client_socket, addr):
        """Thread that triggers event from packet recv from clients"""
        while self.is_online:
            try:
                raw_len = Packet.recv_exact(client_socket, 4)  # Get the length of the packet
                length = struct.unpack("!I", raw_len)[0]
                packet = Packet.recv_exact(client_socket, length)
                if packet:
                    packet_name, contents = Packet.decode_packet(packet)

                    if packet_name == "client_disconnection":
                        server_event_registry.trigger(packet_name, client_socket, *contents, threading_event=False)
                        break
                    else:
                        server_event_registry.trigger(packet_name, client_socket, *contents)
                else:
                    Server.print_server(
                        f"A packet from: {addr} has been send with no data ! This is illegal closing the listening thread and the communication !")
                    break
            except ConnectionResetError:
                server_event_registry.trigger("client_disconnection", client_socket, threading_event=False)
                Server.print_server(f"The client at address: {addr} has been disconnected ! This is an anomaly.")
                break
            except Exception as e:
                print(f"THIS IS NOT NORMAL: {e}")

        if addr in self.clients:
            self.clients.pop(addr)
        client_socket.close()
        Server.print_server(f"The client: {addr} is disconnect !")

    def send_to_all(self, packet_name, *contents):
        """
        Send a packet to all player
        :param packet_name: The name of the packet/event
        :param contents: Contents
        """
        for client in self.clients.values():
            self.send_to(client, packet_name, *contents)

    @staticmethod
    def send_to(clt: socket.socket, packet_name, *contents):
        packet = Packet.create_packet(packet_name, *contents)
        length = len(packet)
        header = struct.pack('!I', length)
        try:
            clt.sendall(header + packet)
        except BrokenPipeError:
            warn(f"The client was disconnect during a sending operation: {clt.getpeername()}")
        except OSError:
            warn("The server is certainly closed...")
        except ConnectionResetError:
            warn(f"A client was disconnect during a sending operation!")


# Basic Event Registry

# SERVER EVENTS
@server_event_registry.on_event("client_disconnection")
def on_client_disconnect(clt, *args):
    Server.print_server(f"The client: {clt.getpeername()} disconnect itself ! (Client Side)")


@server_event_registry.on_event("client_connection")
def on_client_connect(clt, *args):
    Server.print_server(f"Client connected on: {clt.getpeername()} !")


# CLIENT EVENTS
@client_event_registry.on_event("connection")
def on_connection(clt, *args):
    Client.print_client("Client is connecting !")


@client_event_registry.on_event("disconnection")
def on_disconnection(clt, *args):
    Client.print_client("Client disconnecting.")



import pygame
from pygame import Vector2, Surface

from pyreverb import reverb
from pyreverb.Exemple.shooter_objects import Player, Bullet, MAP_SIZE, TICK
from pyreverb.reverb import *

clock = pygame.time.Clock()
reverb.VERBOSE = 2  # make it speak less


@client_event_registry.on_event("server_stop")
def test():
    print("STOP")


def start_client(is_host=False, port=8080, admin_key: int = 1000):
    pygame.init()
    screen: Surface = pygame.display.set_mode(MAP_SIZE)
    is_running = True
    print("Pygame is init !")

    ReverbManager.IS_HOST = is_host
    ReverbManager.REVERB_SIDE = ReverbSide.CLIENT
    clt = Client(port=port)
    ReverbManager.REVERB_CONNECTION = clt
    clt.connect()
    ReverbManager.log_as_admin(admin_key)

    while is_running and ReverbManager.REVERB_CONNECTION.is_connected:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                is_running = False

        screen.fill("purple")

        for p in ReverbManager.get_all_ro_by_type(Player):
            pygame.draw.circle(screen, p.color.get(), tuple(p.pos.get()), 3)

        for b in ReverbManager.get_all_ro_by_type(Bullet):
            pygame.draw.line(screen, b.color.get(), Vector2(b.pos.get()) - Vector2(b.dir.get()),
                             Vector2(b.pos.get()) + Vector2(b.dir.get()), 1)

        pygame.display.flip()
        clock.tick(TICK)

    print("Closing the game...")
    if ReverbManager.IS_HOST:
        ReverbManager.stop_server_admin()
        pygame.quit()
    else:
        clt.disconnect()
    # Stop distant server


if __name__ == "__main__":
    print("Starting client from main!")
    start_client()



class ReverbWrongSideError(Exception):
    def __init__(self, side):
        super().__init__(f"Wrong side: '{side}' side.")


class ReverbUIDAlreadyInitError(Exception):
    def __init__(self, ro, uid):
        super().__init__(
            f"The ReverbObject '{ro.__class__.__name__}' is already init with {uid=} already taken! Only new ReverbObject can have a new uid!")


class ReverbUIDUnknownError(Exception):
    def __init__(self):
        super().__init__(f"The uid is 'Unknown'! This should'nt be 'Unknown'!")


class ReverbObjectAlreadyExistError(Exception):
    def __init__(self, ro):
        super().__init__(f"The ReverbObject '{ro.__class__.__name__}' is already into the ReverbManager!")


class ReverbObjectNotFoundError(Exception):
    def __init__(self, uid):
        super().__init__(f"ReverbObject not found with {uid=}")


class ReverbTypeNotFoundError(Exception):
    def __init__(self, t):
        super().__init__(f"The type={t} is not found into the registry!")



import os
import time

from pyreverb import reverb
from pyreverb.Exemple.client import start_client
from pyreverb.reverb import start_distant, ReverbSide

PORT = 8080
ADMIN_KEY = 1001

if __name__ == "__main__":
    work_dir = os.path.dirname(os.path.abspath(__file__))
    print(work_dir)
    choice = input("Is the host? (Y|n)>>>")
    is_host = False
    pid = ""
    if choice == "Y" or choice == "":
        start_distant(work_dir + "/server.py", str(PORT), str(ADMIN_KEY))
        print("Starting client in 1 second...")
        time.sleep(1)
        is_host = True

    start_client(is_host, PORT, ADMIN_KEY)
